import os
import shlex
from subprocess import Popen, PIPE
from yatr import __version__ as yver
from yatr.base import get_output
from syn.base_utils import chdir

DIR = os.path.abspath(os.path.dirname(__file__))
TEST2 = os.path.join(DIR, 'test2.yml')

#-------------------------------------------------------------------------------

def test_invocation():
    p = Popen(shlex.split('yatr --version'), stdout=PIPE)
    out = p.communicate()[0].decode('utf-8').strip()
    assert out == 'yatr {}'.format(yver)
    assert p.returncode == 0

    out, code = get_output('yatr -f {} bad'.format(TEST2))
    assert out == ''
    assert code == 1

    out, code = get_output('yatr -f {} good'.format(TEST2))
    assert out == ''
    assert code == 0

    # Verify example
    with chdir(os.path.join(DIR, 'example')):
        # Test --dump-path
        out, code = get_output('yatr --dump-path')
        assert out == '{}\n'.format(os.path.join(DIR, 'example/yatrfile.yml'))
        assert code == 0

        # Test nested include
        out, code = get_output('yatr -f C.yml --dump')
        assert 'a = baz\nb = ghi\nc = xyz\n' in out

        # Test included task
        out, code = get_output('yatr foo')
        assert out == 'bar\n'

        # Test default task
        out, code = get_output('yatr')
        assert out == 'bar\n'

        # Test example tasks
        out, code = get_output('yatr cwd')
        assert out == os.getcwd() + '\n'

        out, code = get_output('yatr bar')
        assert out == 'bar\nbar baz xyz\n'

        out, code = get_output('yatr bar foo')
        assert out == 'bar\nbar baz foo\n'

        out, code = get_output('yatr -v bar foo')
        assert out == 'echo bar\nbar\necho bar baz foo\nbar baz foo\n'
        
        out, code = get_output('yatr cond1')
        assert out.strip().split() == ['A.yml',
                                       'B.yml',
                                       'C.yml',
                                       'D.yml',
                                       'yatrfile.yml']

        out, code = get_output('yatr cond2')
        assert out == ''

        out, code = get_output('yatr cond3')
        assert out == ''

        out, code = get_output('yatr cond4')
        assert out == 'bar\n'

        out, code = get_output('yatr -v cond4')
        assert out == 'false\necho bar\nbar\n'

        # Test settings.silent
        out, code = get_output('yatr -f D.yml foo')
        assert out == ''
        assert code == 0

        out, code = get_output('yatr -f D.yml -s silent=false foo')
        assert out == 'bar\n'

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



from syn.base import Base, Attr, create_hook
from syn.base_utils import message
from syn.type import Dict
from syn.five import STR

from .base import ValidationError, eprint

#-------------------------------------------------------------------------------

CONTEXT_REGISTRY = {}

#-------------------------------------------------------------------------------
# Context


class Context(Base):
    context_name = None
    required_opts = ()

    _attrs = dict(inside = Attr(STR, ''),
                  envvars = Attr(Dict(STR), init=lambda self: dict()),
                  opts = Attr(dict, init=lambda self: dict()),
                  _skip_validation = Attr(bool, False, internal=True))
    _opts = dict(init_validate = True)

    @classmethod
    @create_hook
    def _register(cls):
        if cls.context_name:
            CONTEXT_REGISTRY[cls.context_name] = cls

    @classmethod
    def from_yaml(cls, name, dct):
        kwargs = {}
        kwargs['envvars'] = dct.get('env', {})
        kwargs['opts'] = dct.get('opts', {})
        kwargs['inside'] = dct.get('inside', '')

        # TODO: if invalid context specified, raise ValidationError
        cls_ = CONTEXT_REGISTRY.get(dct.get('instanceof', 'null'), cls)
        return cls_(**kwargs)

    def resolve_macros(self, env, **kwargs):
        envvars = dict(self.envvars)
        for key , value in self.envvars.items():
            envvars[key] = env.resolve(value)

        opts = dict(self.opts)
        for key, value in self.opts.items():
            if isinstance(value, STR):
                opts[key] = env.resolve(value)

        return envvars, opts

    def run_command(self, command, env, **kwargs):
        if self.inside:
            ctx = env.contexts[self.inside]
            command = ctx.run_command(command, env, **kwargs)
        return command

    run = None

    def validate(self):
        super(Context, self).validate()

        if self._skip_validation:
            return

        for opt in self.required_opts:
            if opt not in self.opts:
                raise ValidationError('Required option "{}" not defined'
                                      .format(opt))


#-------------------------------------------------------------------------------
# Builtins

#-----------------------------------------------------------
# Null


class Null(Context):
    context_name = 'null'


#-----------------------------------------------------------
# Bash


class Bash(Context):
    context_name = 'bash'

    def run_command(self, command, env, **kwargs):
        # TODO: quote string here
        cmd = 'bash -c "{}"'.format(command)
        return super(Bash, self).run_command(cmd, env, **kwargs)


#-----------------------------------------------------------
# CD


class CD(Context):
    context_name = 'cd'


#-----------------------------------------------------------
# Docker


class Docker(Context):
    context_name = 'docker'


#-----------------------------------------------------------
# Python


# TODO: separate contexts for python-eval and python-exec
class Python(Context):
    context_name = 'python'

    def run(self, command, env, **kwargs):
        try:
            eval(command)
            return 0
        except Exception as e:
            eprint(message(e))
            return 1


#-----------------------------------------------------------
# SSH


class SSH(Context):
    context_name = 'ssh'
    required_opts = ('user', 'hostname')

    def run_command(self, command, env, **kwargs):
        cmd = 'ssh {}@{} "{}"'.format(self.opts['user'],
                                      self.opts['hostname'],
                                      command)
        return super(SSH, self).run_command(cmd, env, **kwargs)


#-----------------------------------------------------------

BUILTIN_CONTEXTS = dict(null = Null(),
                        bash = Bash(),
                        cd = CD(),
                        docker = Docker(),
                        python = Python(),
                        ssh = SSH(_skip_validation=True)) # TODO: need a better way of handling this

#-------------------------------------------------------------------------------
# __all__

__all__ = ('Context',)

#-------------------------------------------------------------------------------



from nose.tools import assert_raises
from syn.base_utils import capture

from yatr import Context, Env, ValidationError
from yatr.context import Null, Bash, Python

#-------------------------------------------------------------------------------
# Context

def test_context():
    env = Env()
    c = Context.from_yaml('foo', {})
    assert type(c) is Null
    
    assert c.run_command('ls', env) == 'ls'

#-------------------------------------------------------------------------------
# Bash

def test_bash():
    env = Env()
    b = Context.from_yaml('foo', dict(instanceof='bash',
                                      env=dict(a='1', 
                                               b='{{bar}}'),
                                      opts=dict(v=True, 
                                                foo='{{a}}')))
    assert type(b) is Bash

    assert b.envvars == dict(a='1', b='{{bar}}')
    assert b.opts == dict(v=True, foo='{{a}}')
    envvars, opts = b.resolve_macros(Env(env=dict(a='4', bar='foo')))
    assert envvars == dict(a='1', b='foo')
    assert opts == dict(v=True, foo='4')
    
    assert b.run_command('ls', env) == 'bash -c "ls"'

#-------------------------------------------------------------------------------
# Python

def test_python():
    env = Env()
    p = Context.from_yaml('foo', dict(instanceof='python'))
    assert type(p) is Python

    assert p.run_command('1/1', env) == '1/1'
    assert p.run('1/1', env) == 0

    with capture() as (out, err):
        assert p.run('1/0', env) == 1
    assert out.getvalue() == ''
    assert 'zero' in err.getvalue()

#-------------------------------------------------------------------------------
# SSH

def test_ssh():
    env = Env()
    assert_raises(ValidationError, Context.from_yaml,
                  'foo', dict(instanceof='ssh'))
                                      
    s = Context.from_yaml('foo', dict(instanceof='ssh',
                                      opts=dict(user='foo',
                                                hostname='bar')))
    assert s.run_command('ls', env) == 'ssh foo@bar "ls"'

    env.contexts['foo'] = s
    sb = Context.from_yaml('bar', dict(instanceof='bash',
                                       inside='foo'))
    assert sb.run_command('ls', env) == 'ssh foo@bar "bash -c \"ls\""'

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



import os
import imp
import yaml
from syn.base_utils import chdir
from syn.base import Base, Attr, init_hook
from syn.type import List, Dict
from syn.five import STR

from .base import ValidationError, resolve_url, resolve, ordered_macros,\
    DEFAULT_CACHE_DIR, str_to_bool, get_delete
from .context import Context
from .task import Task
from .env import Env

STRList = List(STR)

#-------------------------------------------------------------------------------
# Document

class Document(Base):
    _attrs = dict(imports = Attr(List(STR), init=lambda self:list()),
                  includes = Attr(List(STR), init=lambda self: list()),
                  secrets = Attr(List(STR), init=lambda self: list()),
                  macros = Attr(Dict((STR, int, float, 
                                      List((STR, int, float, list, dict)),
                                      Dict((STR, int, float, list, dict)))), 
                                init=lambda self: dict()),
                  contexts = Attr(Dict(Context), init=lambda self: dict()),
                  tasks = Attr(Dict(Task), init=lambda self: dict()),
                  secret_values = Attr(Dict(STR), init=lambda self: dict()),
                  captures = Attr(Dict(STR), init=lambda self: dict()),
                  files = Attr(Dict(STR), init=lambda self: dict()),
                  settings = Attr(Dict(None), init=lambda self: dict()),
                  default_task = Attr(STR, '', 'Task to run if no task is '
                                      'specified at the command line'),
                  env = Attr(Env, init=lambda self: Env(), internal=True),
                  dirname = Attr(STR, doc='Relative path for includes'),
                  cachedir = Attr(STR, '', 'Directory to store downloaded files'),
                  pull = Attr(bool, False, 'Force-pull URLs'),
                 )
    _opts = dict(init_validate = True)

    @classmethod
    def from_path(cls, path, **kwargs):
        with open(path, 'r') as f:
            dct = yaml.load(f)
            return cls.from_yaml(dct,
                                 os.path.abspath(os.path.dirname(path)), 
                                 **kwargs)

    @classmethod
    def from_yaml(cls, dct, dirname, **kwargs):
        get_delete(dct, kwargs, 'import', [], 'imports')
        get_delete(dct, kwargs, 'include', [], 'includes')
        get_delete(dct, kwargs, 'capture', {}, 'captures')
        get_delete(dct, kwargs, 'files', {})
        get_delete(dct, kwargs, 'secrets', [])
        get_delete(dct, kwargs, 'macros', {})
        get_delete(dct, kwargs, 'contexts', {})
        get_delete(dct, kwargs, 'tasks', {})
        get_delete(dct, kwargs, 'default', '', 'default_task')

        init_macros = kwargs.get('initial_macros', {})
        for name, value in init_macros.items():
            if name not in kwargs['macros']:
                kwargs['macros'][name] = value

        settings = dict(dct.get('settings', {}))
        settings.update(kwargs.get('settings', {}))
        kwargs['settings'] = settings
        if 'settings' in dct:
            del dct['settings']

        if dct:
            raise ValidationError('Invalid top-level keys: {}'
                                  .format(','.join(dct.keys())))

        for key in list(kwargs['contexts'].keys()):
            kwargs['contexts'][key] = \
                Context.from_yaml(key, kwargs['contexts'][key])

        for key in list(kwargs['tasks'].keys()):
            kwargs['tasks'][key] = \
                Task.from_yaml(key, kwargs['tasks'][key])

        kwargs['dirname'] = dirname
        return cls(**kwargs)

    @init_hook
    def process(self, **kwargs):
        cachedir = DEFAULT_CACHE_DIR
        if self.cachedir:
            cachedir = self.cachedir

        self.process_settings(**kwargs)

        jenv = Env().jenv
        pre_macros = dict(self.macros)
        for name, macro in ordered_macros(pre_macros, lenient=True):
            try:
                pre_macros[name] = resolve(macro, pre_macros, lenient=True,
                                           jenv=jenv)
            except:
                pass # There might be macros defined in terms of
                     # jinja_functions to be imported

        def process(path):
            return resolve_url(resolve(path, pre_macros, jenv=jenv), 
                               cachedir=cachedir, force=self.pull)

        with chdir(self.dirname):
            for path in map(process, self.imports):
                if not os.path.exists(path):
                    raise ValidationError("Module path does not exist: {}"
                                          .format(path))
                self.process_import(path, **kwargs)

            for path in map(process, self.includes):
                if not os.path.isfile(path):
                    raise ValidationError("Include path does not exist: {}"
                                          .format(path))
                self.process_include(path, **kwargs)

            files = {}
            for fname, fpath in self.files.items():
                path = process(fpath)
                if not os.path.isfile(path):
                    raise ValidationError("File path does not exist: {}"
                                          .format(path))
                files[fname] = path

        for name in self.secrets:
            self.process_secret(name, **kwargs)

        env = Env(macros=self.macros, contexts=self.contexts, tasks=self.tasks,
                  secret_values=self.secret_values, captures=self.captures,
                  settings=self.settings, default_task=self.default_task,
                  files=files)
        self.env.update(env, **kwargs)

    def post_process(self, **kwargs):
        with chdir(self.dirname):
            self.env.resolve_macros()
            self.validate()

    def process_import(self, path, **kwargs):
        mod = imp.load_source('yatr_module_import', path)
        
        if not hasattr(mod, 'env'):
            raise ImportError("yatr extension module '{}' has no env"
                              .format(path))

        self.env.update(mod.env, **kwargs)

    def process_include(self, path, **kwargs):
        doc = Document.from_path(path, pull=self.pull, cachedir=self.cachedir,
                                 settings=self.settings)
        self.env.update(doc.env, **kwargs)

    def process_secret(self, name, **kwargs):
        raise NotImplementedError('Secrets currently unsupported')

    def process_settings(self, **kwargs):
        self.settings['silent'] = \
            str_to_bool(self.settings.get('silent', False))
        self.settings['loop_count_macro'] = \
            self.settings.get('loop_count_macro', '_n')
        self.settings['preview_conditionals'] = \
            str_to_bool(self.settings.get('preview_conditionals', True))
        self.settings['exit_on_error'] = \
            str_to_bool(self.settings.get('exit_on_error', True))

    def run(self, name, **kwargs):
        try:
            with chdir(self.dirname):
                return self.env.tasks[name].run(self.env, **kwargs)
        except KeyError:
            raise RuntimeError('No such task: {}'.format(name))

    def validate(self):
        super(Document, self).validate()
        self.env.validate()


#-------------------------------------------------------------------------------

DEFAULT_SETTINGS = sorted(Document(dirname=os.getcwd()).env.settings.keys())

#-------------------------------------------------------------------------------
# __all__

__all__ = ('Document',)

#-------------------------------------------------------------------------------



import os
from yatr import Document
from yatr.base import resolve_url

DIR = os.path.abspath(os.path.dirname(__file__))
YF1 = os.path.join(DIR, 'yatrfile1.yml')
URL = 'https://raw.githubusercontent.com/mbodenhamer/yatrfiles/master/yatrfiles/test/test1.yml'

#-------------------------------------------------------------------------------
# Utilities

def test_resolve_url():
    assert resolve_url(YF1) == YF1
    Document.from_path(YF1).post_process()

    path = resolve_url(URL)
    assert path.startswith(os.path.expanduser('~'))
    Document.from_path(path).post_process()

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)



from yatr import Env

def foo(value, **kwargs):
    return '{}_foo'.format(value)

def bar(value, **kwargs):
    return '{}_bar'.format(value)

env = Env(jinja_filters=dict(foo=foo),
          jinja_functions=dict(bar=bar))



import os
import re
import sys
import json
import shutil
import hashlib
from argparse import ArgumentParser

from . import __version__ as yver
from .base import DEFAULT_CACHE_DIR, cached_path, resolve
from .parse import DEFAULT_SETTINGS, Document

DIR = os.path.abspath(os.path.dirname(__file__))

#-------------------------------------------------------------------------------

OPTION_STRINGS = []

def add_argument(parser, *args, **kwargs):
    opt = parser.add_argument(*args, **kwargs)
    OPTION_STRINGS.extend(opt.option_strings)

#-------------------------------------------------------------------------------

DESCRIPTION = 'Yet Another Task Runner.'

parser = ArgumentParser(prog='yatr', description=DESCRIPTION)

# Value options
add_argument(parser, '-f', '--yatrfile', dest='yatrfile', type=str,
             default='', metavar='<yatrfile>', help='The yatrfile to load')
add_argument(parser, '-i', dest='infile', type=str, default='',
             metavar='<file>', help='Input file')
add_argument(parser, '-o', dest='outfile', type=str, default='',
             metavar='<file>', help='Output file')
add_argument(parser, '-m', '--macro', dest='macros',
             action='append', metavar='<macro>=<value>',
             help='Set/override macro with specified value')
add_argument(parser, '-s', '--setting', dest='settings',
             action='append', metavar='<setting>=<value>',
             help='Set/override setting with specified value')
add_argument(parser, '--cache-dir', dest='cachedir', type=str,
             default=DEFAULT_CACHE_DIR, metavar='<DIR>',
             help='Path of cache directory')

# Switches
add_argument(parser, '-v', '--verbose', dest='verbose', default=False,
             action='store_true', help='Print commands to be run')
add_argument(parser, '-p', '--preview', dest='preview', default=False,
             action='store_true', help='Preview commands to be run '
             'without running them (implies -v)')

# Commands
add_argument(parser, '--cache', dest='cache_file', default=False,
             action='store_true', help='Cache local input file (-i) as if it '
             'were a URL (-o)')
add_argument(parser, '--dump', dest='dump_vars', default=False,
             action='store_true', help='Dump macro values and exit')
add_argument(parser, '--dump-path', dest='dump_path', default=False,
             action='store_true', help='Print yatrfile path and exit')
add_argument(parser, '--pull', dest='pull', default=False, action='store_true',
             help='Download all URL includes and imports, then exit')
add_argument(parser, '--render', dest='render', default=False, 
             action='store_true', help='Use macros to render a Jinja2 template'
             ' file (requires -i and -o)')
add_argument(parser, '--version', dest='show_version', default=False,
             action='store_true', help='Print version info and exit')
add_argument(parser, '--validate', dest='validate', default=False,
             action='store_true', help='Validate the yatrfile and exit')
add_argument(parser, '--install-bash-completions', default=False,
             dest='install_bash_completions', action='store_true',
             help='Install bash tab completion script globally, then exit')

# Positional args
add_argument(parser, 'task', metavar='<task>', type=str, default='', nargs='?',
             help='The task to run')
add_argument(parser, 'args', metavar='ARGS', nargs='*',
             help='Additional arguments for the task')

USAGE = parser.format_usage().strip()
OPTION_STRINGS.sort()

#-------------------------------------------------------------------------------

def print_(s, flush=True):
    sys.stdout.write(s + '\n')
    if flush:
        sys.stdout.flush()

#-------------------------------------------------------------------------------
# Yatrfile search

YATRFILE_PATTERN = re.compile('^[Yy]atrfile(.yml)?$')

def search_dir(path):
    for name in os.listdir(path):
        if re.match(YATRFILE_PATTERN, name):
            if os.path.isfile(os.path.join(path, name)):
                return os.path.join(path, name)

def search_rootward(path):
    res = search_dir(path)

    if res:
        return res
    if path == '/':
        raise RuntimeError('Unable to find yatrfile')
        
    return search_rootward(os.path.dirname(path))

def find_yatrfile_path(path):
    if path:
        return path
    return search_rootward(os.path.abspath(os.getcwd()))

#-------------------------------------------------------------------------------
# Completion data

def default_data():
    ret = {}
    ret['tasks'] = []
    ret['macros'] = []
    ret['settings'] = DEFAULT_SETTINGS
    return ret

def data_path_from_yatrfile_path(path, cachedir):
    base = hashlib.sha224(path.encode('utf-8')).hexdigest()
    fname = base + '_compdata'
    fpath = os.path.join(cachedir, fname)
    return fpath

def compile_completion_data(path, cachedir, outpath):
    data = default_data()
    try:
        doc = Document.from_path(path, cachedir=cachedir)
    except:
        return data

    data['tasks'] = sorted(doc.env.tasks.keys())
    data['macros'] = sorted(doc.env.macros.keys())
    data['settings'] = sorted(doc.env.settings.keys())

    if not os.path.isdir(cachedir):
        os.mkdir(cachedir)

    with open(outpath, 'w') as f:
        json.dump(data, f)

    return data

def load_completion_data(yatrfile, cachedir):
    cachedir = os.path.expanduser(cachedir)
    try:
        path = find_yatrfile_path(yatrfile)
    except:
        return default_data()

    dpath = data_path_from_yatrfile_path(path, cachedir)

    if not os.path.isfile(dpath):
        data = compile_completion_data(path, cachedir, dpath)

    elif os.stat(dpath).st_mtime <= os.stat(path).st_mtime:
        data = compile_completion_data(path, cachedir, dpath)

    else:
        with open(dpath, 'r') as f:
            data = json.load(f)

    return data

#-------------------------------------------------------------------------------
# Bash completions

def matches(s, lst):
    return [elem for elem in lst if elem.startswith(s)]

def find_bash_completions(args, idx):
    def find_opt(name, default):
        if name in args:
            n_idx = args.index(name)
            if n_idx == idx - 1:
                return # We are trying to complete a pathname
            elif n_idx < len(args) - 1:
                return args[n_idx + 1]
        return default

    cachedir = find_opt('--cache-dir', DEFAULT_CACHE_DIR)
    if cachedir is None:
        return [] # we are completing directory path; let defaults handle it

    yfpath = find_opt('-f', '')
    if yfpath is None:
        return [] # we are completing a file path; let defaults handle it

    # if query on '-f' returns '', check if '--yatrfile' was given
    if not yfpath:
        yfpath = find_opt('--yatrfile', '')
        if yfpath is None:
            return [] # we are completing a file path; let defaults handle it

    data = load_completion_data(yfpath, cachedir)

    if idx < len(args): # we are completing an already-started word
        word = args[idx]
        if word.startswith('-'):
            return matches(word, OPTION_STRINGS)

        elif idx > 0:
            pword = args[idx - 1]
            if any(w in data['tasks'] for w in args):
                return [] # we are at a task argument; let defaults handle it

            elif pword in ('-m', '--macro'):
                if '=' in word:
                    return [] # we are completing a macro value
                else:
                    return matches(word, data['macros'])

            elif pword in ('-s', '--setting'):
                if '=' in word:
                    return [] # we are completing a setting value
                else:
                    return matches(word, data['settings'])

            elif pword in ('-i', '-o'):
                return [] # we are completing a filename

            else: # The only other option is that we are completing a task name
                return matches(word, data['tasks'])

        else: # The first non-'-'-prefixed argument; must be a task name
            return matches(word, data['tasks'])
    
    else: # we are starting a new word
        if idx == 0: # probably looking for a specific task, so return all names
            return data['tasks']
        
        else:
            pword = args[idx - 1]
            if any(w in data['tasks'] for w in args):
                return [] # we are at a task argument; job for defaults
            
            elif pword in ('-m', '--macro'):
                return data['macros']
            
            elif pword in ('-s', '--setting'):
                return data['settings']

            elif pword in ('-i', '-o'):
                return [] # we are completing a filename

            else: # probably looking for a task at this point
                return data['tasks']

def dump_bash_completions(args, idx):
    comps = find_bash_completions(args, idx)
    out = ' '.join(comps)
    print(out)

BASH_COMPLETION_MESSAGE = \
'''If custom tab completions are not enabled globally, you may need to add the following line to ~/.bashrc:

    source /etc/bash_completion.d/yatr

which can be accomplished by running the following:

    echo "source /etc/bash_completion.d/yatr" >> ~/.bashrc
'''

def install_bash_completions():
    script = os.path.join(DIR, 'scripts/completion/yatr')
    shutil.copyfile(script, '/etc/bash_completion.d/yatr')
    print(BASH_COMPLETION_MESSAGE)

#-------------------------------------------------------------------------------
# --cache

def cache_file(infile, outfile, cachedir):
    path = cached_path(outfile, cachedir)
    shutil.copyfile(infile, path)

#-------------------------------------------------------------------------------
# --render

def render(doc, infile, outfile):
    with open(infile, 'r') as f:
        template = f.read()

    out = resolve(template, doc.env.env, jenv=doc.env.jenv)
    
    with open(outfile, 'w') as f:
        f.write(out)

#-------------------------------------------------------------------------------
# Main

def _main(*args):
    if args:
        if args[0] == '--dump-bash-completions':
            dump_bash_completions(args[2:], int(args[1]))
            return

    opts = parser.parse_args(args)
    
    if opts.install_bash_completions:
        install_bash_completions()
        return

    if opts.cache_file:
        if not opts.infile:
            raise RuntimeError('Required option -i not given')
        if not opts.outfile:
            raise RuntimeError('Required option -o not given')

        cache_file(opts.infile, opts.outfile, opts.cachedir)
        return

    # --version
    if opts.show_version:
        print('yatr {}'.format(yver))
        return

    # Validate -i
    if opts.infile:
        if not os.path.isfile(opts.infile):
            raise RuntimeError('Input file "{}" does not exist'
                               .format(opts.infile))

    # Populate position arg macros
    init_macros = {}
    init_macros['ARGS'] = list(opts.args)
    for k, arg in enumerate(opts.args):
        key = '_{}'.format(k+1)
        init_macros[key] = arg

    # Process settings
    settings = {}
    if opts.settings:
        for s in opts.settings:
            setting, value = s.split('=')
            settings[setting] = value

    # Load yatrfile
    path = find_yatrfile_path(opts.yatrfile)
    init_macros['YATRFILE'] = path
    init_macros['YATR'] = "yatr -f {{YATRFILE}}"
    init_macros['CURDIR'] = os.path.abspath(os.path.dirname(path))
    doc = Document.from_path(path, pull=opts.pull, cachedir=opts.cachedir,
                             settings=settings, initial_macros=init_macros)
    
    # If we executed --pull, then exit
    if opts.pull:
        return

    # Process command-line macro overrides
    if opts.macros:
        for s in opts.macros:
            macro, value = s.split('=')
            doc.env.macros[macro] = value
    
    doc.post_process()

    # --dump-path
    if opts.dump_path:
        print_(path)
        return

    # --dump
    if opts.dump_vars:
        # TODO: add support for filtering out unwanted variables
        # TODO: add support for not including possible secrets in output
        displayable = (set(doc.env.macros) | 
                       set(doc.env.captures) | 
                       set(doc.env.files))
        for name in sorted(doc.env.env):
            if name in displayable:
                print_('{} = {}'.format(name, doc.env.env[name]))
        return

    # --validate
    if opts.validate:
        # Check that there are no undefined macros in task definitions
        for task in doc.env.tasks.values():
            task.run_preview(doc.env)

        print("Validation successful")
        return

    # --render
    if opts.render:
        if not opts.infile:
            raise RuntimeError('Required option -i not given')
        if not opts.outfile:
            raise RuntimeError('Required option -o not given')

        render(doc, opts.infile, opts.outfile)
        return

    if opts.preview:
        opts.verbose = True

    task = opts.task
    if not task and doc.env.default_task:
        task = doc.env.default_task

    if task:
        codes = doc.run(task, preview=opts.preview, verbose=opts.verbose)
        if codes:
            sys.exit(max(codes))

    elif not args:
        print(USAGE)
        sys.exit(1)

#-------------------------------------------------------------------------------

def main():
    _main(*sys.argv[1:])

#-------------------------------------------------------------------------------



from setuptools import setup

def read(fpath):
    with open(fpath, 'r') as f:
        return f.read()

def requirements(fpath):
    return list(filter(bool, read(fpath).split('\n')))

def version(fpath):
    return read(fpath).strip()

setup(
    name = 'yatr',
    version = version('version.txt'),
    author = 'Matt Bodenhamer',
    author_email = 'mbodenhamer@mbodenhamer.com',
    description = 'Yet Another Task Runner',
    long_description = read('README.rst'),
    url = 'https://github.com/mbodenhamer/yatr',
    packages = ['yatr'],
    package_dir = {'yatr': 'yatr'},
    package_data = {'yatr': 'scripts'},
    include_package_data = True,
    install_requires = requirements('requirements.in'),
    entry_points = {
        'console_scripts': [
            'yatr = yatr.main:main',
        ]
    },
    license = 'MIT',
    keywords = ['task', 'make', 'yaml'],
    classifiers = [
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: MIT License',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.6',
        'Topic :: Utilities'
    ]
)



import os
from nose.tools import assert_raises

import yatr.context as yc
from yatr import Document, ValidationError
from yatr.parse import get_delete

DIR = os.path.abspath(os.path.dirname(__file__))

#-------------------------------------------------------------------------------
# Utilities

def test_get_delete():
    d1 = dict(a=1, b=2)
    d2 = {}
    get_delete(d1, d2, 'a', 3)
    assert d1 == dict(b=2)
    assert d2 == dict(a=1)

    get_delete(d1, d2, 'b', 3, 'd')
    assert d1 == {}
    assert d2 == dict(a=1, d=2)

    get_delete(d1, d2, 'c', 3)
    assert d1 == {}
    assert d2 == dict(a=1, d=2, c=3)

#-------------------------------------------------------------------------------
# Document

def test_document():
    d = Document(dirname=DIR)
    assert not d.tasks

    d = Document.from_yaml({}, DIR)
    assert not d.tasks

    d = Document.from_yaml(dict(contexts=dict(bash1=dict(instanceof='bash')),
                                tasks=dict(dir='pwd')),
                           DIR)
    codes = d.run('dir', silent=True)
    assert codes == [0]

    d = Document.from_yaml(dict(capture=dict(cwd='pwd')), os.getcwd())
    d.post_process()
    assert d.env.env['cwd'] == os.getcwd()

    path = os.path.join(DIR, 'mod1.py')
    assert_raises(ImportError, Document.from_yaml, {'import': [path]}, DIR)

    assert 'foo' not in yc.CONTEXT_REGISTRY
    path = os.path.join(DIR, 'mod2.py')
    d = Document.from_yaml({'import': [path]}, DIR)
    assert 'foo' in d.env.contexts
    assert 'foo' in yc.CONTEXT_REGISTRY

    path = os.path.join(DIR, 'yatrfile1.yml')
    d = Document.from_yaml(dict(include=[path],
                                macros=dict(a=3)),
                           DIR)
    assert d.env.macros == dict(a=3, b=2)
    assert_raises(RuntimeError, d.run, 'foobarbaz')

    assert_raises(ValidationError, Document.from_yaml, dict(foo='a'), DIR)
    assert_raises(ValidationError, Document.from_yaml, dict(include=['/foo/bar']), DIR)
    assert_raises(ValidationError, Document.from_yaml, {'import': ['/foo/bar']}, DIR)
    assert_raises(ValidationError, Document.from_yaml, {'files': {'a': '/foo/bar'}}, DIR)
    assert_raises(NotImplementedError, Document.from_yaml, dict(secrets=['a']), DIR)

    d = Document.from_yaml(dict(macros=dict(a='{{b}}')), os.getcwd())
    assert_raises(ValidationError, d.env.resolve_macros)

#-------------------------------------------------------------------------------

if __name__ == '__main__': # pragma: no cover
    from syn.base_utils import run_all_tests
    run_all_tests(globals(), verbose=True, print_errors=False)
